/*
 * wheel.c
 *
 * Created: 2012/4/17 16:09:57
 *  Author: Andygongyb
 */ 

#include <avr/io.h>
#include <avr/interrupt.h>
#include "wheel.h"

static int16_t wheel_counter = 0x186;
static uint16_t wheel_int_counter = 0;
static uint16_t wheel_pid_int_counter = 0;
static uint8_t wheel_at_break = 0;

static int16_t pid_set = 0;
static int16_t pre_error = 0;
float integral = 0.0;
float pre_integral = 0.0;
float derivative = 0.0;
float kp = 0.13;
float kd = 0.0002;
float ki = 0.05;
#define DT 0.05

void wheel_init(void) {
	cli();
	PORTB &= ~_BV(PB3);
	DDRB |= _BV(PB3);
	PORTD |= _BV(PD2);
	DDRD &= ~_BV(PD2);
	MCUCR |= _BV(ISC01) | _BV(ISC00);
	GICR |= _BV(INT0);
	SET_WHEEL_IN0;
	SET_WHEEL_IN1;
	WHEEL_DDR_IN0 |= _BV(WHEEL_IN0);
	WHEEL_DDR_IN1 |= _BV(WHEEL_IN1);
	TCCR2 = _BV(WGM20) | _BV(WGM21) | _BV(COM21);
	TIMSK |= _BV(TOIE2);
	OCR2 = 255;
	TCCR2 |= _BV(CS21); // period is 0.128ms, 50ms is about 390 periods
	sei();
}

void wheel_set(int16_t speed) {
	if (speed > 255) {
		speed = 255;
	} else if (speed < -255) {
		speed = -255;
	}
	cli();
	pid_set = 0;
	if (speed > 0) {
		SET_WHEEL_IN0;
		CLR_WHEEL_IN1;
	} else {
		SET_WHEEL_IN1;
		CLR_WHEEL_IN0;
		speed = -speed;
	}
	OCR2 = speed;
	sei();
}

void wheel_break(void) {
	cli();
	pid_set = 0;
	SET_WHEEL_IN0;
	SET_WHEEL_IN1;
	OCR2 = 255;
	wheel_at_break = 1;
	sei();
}

void wheel_pid_set(int8_t speed) {
	if (speed == 0) {
		wheel_break();
		return;
	}
	cli();
	if (speed > 10) {
		speed = 10;
	} else if (speed < -10) {
		speed = -10;
	}
	if (speed > 0) {
		pid_set = (10 - speed) * 15 + 100;
		SET_WHEEL_IN0;
		CLR_WHEEL_IN1;
	} else {
		pid_set = -((-10 - speed) * 15 - 100);
		SET_WHEEL_IN1;
		CLR_WHEEL_IN0;
	}
	if (wheel_at_break) {
		wheel_at_break = 0;
		OCR2 = 10;
	}
	pre_error = wheel_counter - pid_set;
	integral = 0.0;
	pre_integral = 0.0;
	derivative = 0;
	sei();
}

uint16_t wheel_get_counter(void) {
	return wheel_counter;
}

ISR(INT0_vect) {
	if (wheel_int_counter > 10) {
		wheel_counter = wheel_int_counter;
		wheel_int_counter = 0;
	}
}

ISR(TIMER2_OVF_vect, ISR_NOBLOCK) {
	if (wheel_int_counter < 0x186) {
		wheel_int_counter++;
	} else {
		wheel_counter = wheel_int_counter;
	}
	
	if (pid_set) {
		wheel_pid_int_counter++;
		// every 50ms
		if (wheel_pid_int_counter == 0x186) {
			wheel_pid_int_counter = 0;
			int16_t diff = wheel_counter - pid_set;
			if ((diff > 1) || (diff < -1)) {
				integral = pre_integral + diff * DT;
				derivative = (diff - pre_error) / DT;
				int16_t pid_out = kp * diff + ki * integral + kd * derivative;
				pre_error = diff;
				pre_integral = integral;
				if (pid_out > 0) {
					pid_out += OCR2;
					if (pid_out > 255) {
						pid_out = 255;
					}
					OCR2 = pid_out;
				} else {
					pid_out += OCR2;
					if (pid_out < 0) {
						pid_out = 0;
					}
					OCR2 = pid_out;
				}
			}			
		}			
	}	
}

