/**
 * \file
 *
 * \brief Empty user application template
 *
 */

/*
 * Include header files for all drivers that have been imported from
 * Atmel Software Framework (ASF).
 */
#include <stdint.h>
#include <avr32/io.h>
#include <stdio.h>

extern "C" {
#include <asf.h>
#include "FreeRTOS/FreeRTOS.h"
#include "FreeRTOS/task.h"
}

#include "DataBlock.hpp"
#include "PDCA.hpp"
#include "SPI.hpp"
#include "OLED_DMA.hpp"
#include "TWI_ASYN.hpp"
#include "Initialization.hpp"
#include "EIC.hpp"
#include "Wireless2.4GHz.hpp"
#include "MPU6050.hpp"
#include "Slave.hpp"
#include "Omni.hpp"
#include "Command.hpp"

void init_task(void* par);
void test_task(void* par);
void send_eular(void* par);
void update(void* par);
void slave_test(void* par);
void balancing(void* par);
void servo_pro_test(void* par);
void receive_cmd(void* par);

xTaskHandle balancing_hdl;

COLED* oled;
CTWIM* twim;
CWireless* the_wireless;
CMPU* the_mpu;
CDataControl* the_data_control;
COmniDir* the_dir;
CCommand* the_cmd;

int main(void) {
	Enable_global_interrupt();
	CInit::instance();
	the_data_control = CDataControl::instance();
	
	xTaskCreate(init_task, (const signed portCHAR *)"init", 2024, NULL, 5, NULL);
	vTaskStartScheduler();	
	return 0;

}

#if (PERFORMANCE_ANALYSIS == 1)
static portTickType last_seen_tick = 0;
static uint32_t idle_max = 0, idle_min = 0;
xTaskHandle last_seen_task = NULL;
xTaskHandle idle_handle = NULL;

void performance_track(void* par);

void performance_track(void* par) {
	portTickType xLastWakeTime = xTaskGetTickCount();
	while(1) {
		static char str[32];
		float tick = xTaskGetTickCount();
		sprintf(str, "%5f %5f\r\n", 1.0 - ((float)idle_max)/tick, 1.0 - ((float)idle_min)/tick);
		the_wireless->append((uint8_t*)str, strlen(str));
		vTaskDelayUntil(&xLastWakeTime, 1000);
	}
}

extern "C" {
void vApplicationTickHook(void) {
	xTaskHandle tmp = xTaskGetCurrentTaskHandle();
	if ((idle_handle == last_seen_task) && (idle_handle == tmp)) {
		idle_min++;
	}
	last_seen_task = tmp;
}
}
#endif

void init_task(void* par) {
	/*CSPI *the_spi =*/
	CSPI::instance();

	CTWIM_ch<CTWIM::TWIM1>::instance();
	oled = COLED::instance();
	twim = CTWIM_ch<CTWIM::TWIM0>::instance();
	the_wireless = CWireless::instance();
	the_mpu = CMPU::instance();
	the_dir = new COmniDir(new CSlave(0x08), new CSlave(0x04), new CSlave(0x02), new CSlave(0x01), the_mpu);
	the_cmd = new CCommand(the_dir);
	xTaskCreate(test_task, (const signed portCHAR *)"test", 1024, NULL, 2, NULL);
#if (PERFORMANCE_ANALYSIS == 1)
	xTaskCreate(performance_track, (const signed portCHAR *)"performance track", 1024, NULL, 2, NULL);
#else
	xTaskCreate(send_eular, (const signed portCHAR *)"send eular", 512, NULL, 2, NULL);
#endif
	xTaskCreate(update, (const signed portCHAR *)"update sensor", 512, NULL, 3, NULL);
	xTaskCreate(balancing, (const signed portCHAR *)"balancing", 512, NULL, 4, &balancing_hdl);
	xTaskCreate(receive_cmd, (const signed portCHAR *)"receive command", 512, NULL, 2, NULL);
	vTaskSuspend(balancing_hdl);
	vTaskDelete(NULL);
}

void send_eular(void* par) {
	portTickType xLastWakeTime = xTaskGetTickCount();
	while(1) {
		float angle[3] = {0.0, 0.0, 0.0};
		the_mpu->get_euler(angle);
		uint8_t buffer[16] = {0, 0, 0, 0};
		memcpy(buffer + 4, angle, 12);
		the_wireless->append(buffer, 16);
		vTaskDelayUntil(&xLastWakeTime, 50);
	}
}

void test_task(void* par) {
	CSlave* slave[4];
	slave[0] = new CSlave(0x01);
	slave[1] = new CSlave(0x02);
	slave[2] = new CSlave(0x04);
	slave[3] = new CSlave(0x08);
	portTickType xLastWakeTime = xTaskGetTickCount();
	while(1) {
		float angle[3] = {0.0, 0.0, 0.0};
		the_mpu->get_euler(angle);
		oled->clear();
		oled->locate(0, 0);
		*oled << "eular angles:";
		for (uint8_t i = 0; i < 3; i++) {
			char str[20] = {0};
			sprintf(str, "%5f", angle[i]);
			oled->locate(0, i + 1);
			*oled << str;
		}
		oled->locate(0, 4);
		*oled << "pos:";
		for (uint8_t i = 0; i < 4; i++) {
			char str[20] = {0};
			sprintf(str, "%u", slave[i]->get_adc());
			oled->locate(i * 6 * 4 + 6 * 5, 4);
			*oled << str;
		}
		oled->locate(0, 5);
		*oled << "spd:";
		for (uint8_t i = 0; i < 4; i++) {
			char str[20] = {0};
			sprintf(str, "%u", slave[i]->get_wheel_speed());
			oled->locate(i * 6 * 4 + 6 * 5, 5);
			*oled << str;
		}
		char str[20];
		the_cmd->copy_cmd(str);
		oled->locate(0, 6);
		*oled << "cmd received:";
		oled->locate(0, 7);
		*oled << str;
		vTaskDelayUntil(&xLastWakeTime, 100);
	}
}

void update(void* par) {
	portTickType xLastWakeTime = xTaskGetTickCount();
	while(1) {
		the_mpu->update_package();
		vTaskDelayUntil(&xLastWakeTime, 20);
	}
}

void balancing(void* par) {
	portTickType xLastWakeTime = xTaskGetTickCount();
	while(1) {
		the_dir->update();
		vTaskDelayUntil(&xLastWakeTime, 50);
	}
}

void receive_cmd(void* par) {
	the_dir->west()->set_servo_pro(8, 54);
	the_dir->east()->set_servo_pro(9, 54);
	the_dir->north()->set_servo_pro(8, 54);
	the_dir->south()->set_servo_pro(8, 56);
	the_dir->west()->set_adc_thr(140, 160);
	the_dir->east()->set_adc_thr(132, 170);
	the_dir->south()->set_adc_thr(155, 210);
	the_dir->north()->set_adc_thr(168, 220);
	the_dir->east()->set_worm(800);
	the_dir->west()->set_worm(800);
	the_dir->north()->set_worm(800);
	the_dir->south()->set_worm(800);
	vTaskDelay(500);
	the_dir->east()->worm_break();
	the_dir->west()->worm_break();
	the_dir->north()->worm_break();
	the_dir->south()->worm_break();
	portTickType xLastWakeTime = xTaskGetTickCount();
	while (1) {
		static uint8_t *buffer;
		if (the_wireless->get_rx_que_len() == 0) {
			the_dir->wheel_break();
		} else {
			while (the_wireless->get_rx_que_len()) {
				the_wireless->get_from_queue(buffer);
				switch (the_cmd->execute_cmd(buffer)) {
				case CCommand::DISABLE_SUSP:
					vTaskSuspend(balancing_hdl);
					the_dir->east()->set_worm(800);
					the_dir->west()->set_worm(800);
					the_dir->north()->set_worm(800);
					the_dir->south()->set_worm(800);
					break;
				case CCommand::ENABLE_SUSP:
					vTaskResume(balancing_hdl);
					break;
				default:
					break;
				}
				delete buffer;
			}
		}

		vTaskDelayUntil(&xLastWakeTime, 100);
	}
}

extern "C" void vApplicationIdleHook(void);

void vApplicationIdleHook(void) {
	vTaskSuspendAll();
	the_data_control->free_all();
#if (PERFORMANCE_ANALYSIS == 1)
	idle_handle = xTaskGetCurrentTaskHandle();
	portENTER_CRITICAL();
	portTickType cur_tick_count = xTaskGetTickCount();
	if (last_seen_tick != cur_tick_count) {
		last_seen_tick = cur_tick_count;
		idle_max++;
	}
	portEXIT_CRITICAL();
#endif
	xTaskResumeAll();
}