/*
 * USB_CDC.hpp
 *
 *  Created on: Nov 27, 2011
 *      Author: Andygongyb
 */

#ifndef USB_HPP_
#define USB_HPP_

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <avr/wdt.h>
#include "Initialization.hpp"
#include "util/delay.h"

extern "C" void USB_GEN_vect(void)__attribute__((signal,used));
extern "C" void USB_COM_vect(void)__attribute__((signal,used));

namespace usb_cdc_names {

const uint8_t device_descriptor[18] = { 0x12, // bLength field, device descriptor has 18 bytes
		0x01, // bDescriptorType field, the device descriptor is 0x01
		// bcdUSB field, we use USB1.1 i.e. 0x0110
		0x10, 0x01, // due to little endian, the actual bytes are 0x10, 0x01
		0x02, // bDeviceClass field, communication class is 0x02
		0x00, // bDeviceSubClass field
		0x00, // bDeviceProtocol field
		0x40, // bMaxPacketSize0 field, end point 0 has 64 bytes capacity
		0xC0, 0x16, // idVender field, we use ATMEL's free VID 0x16C0
		0xDF, 0x05, // idProduct field, we use ATMEL's free PID for CDC class, which is 0x05DF
		0x00, 0x01, // bcdDevice field
		0x01, // iManufacturer field
		0x02, // iProduct field
		0x03, // iSerialNumber field
		0x01 // bNumConfigurations field, we only need one configuration
		};

// configuration descriptor set, totally 67 bytes
const uint8_t conf_descriptor[67] = { 0x09, // bLength field, configuration descriptor has 9 bytes
		0x02, // bDescriptorType field, configuration descriptor is 0x02
		// wTotalLength field, the total length of configuration descriptor set
		// including configuration descriptor, interface descriptor, class descriptor, and end point descriptor
		67 & 0xFF, // low byte
		(67 >> 8) & 0xFF, // high byte
		0x02, // bNumInterfaces field, the number of interfaces in the configuration is 2
		0x01, // bConfigurationValue field
		0x00, // iConfiguration field
		// bmAttributes field, we need the USB bus to power the device, and we do not need remote wake-up, so
		// the value is 0x80
		0x80,
		// bMaxPower field, the maximum power consumption. we will try 200mA
		0x64, // the unit value is 2mA, so we use 100 (0x64)


		0x09, // bLength field, CDC class interface descriptor has 9 bytes
		0x04, // bDescriptorType field, interface descriptor is 0x04
		0x00, // bInterfaceNumber field
		0x00, // bAlternateSetting field
		0x01, // bNumEndpoints field, we use 1 end point other than EP0
		0x02, // bInterfaceClass field, CDC class is 0x02
		0x02, // bInterfaceSubClass field, we use Abstract Control Model, which is 0x02
		0x01, // bInterfaceProtocol field, we use Common AT Commands, which is 0x01
		0x00, // iConfiguration field


		0x05, // bFunctionLength field, Header Functional Descriptor has 5 bytes
		0x24, // bDescriptorType field, CS_INTERFACE is 0x24
		0x00, // bDescriptorSubtype field, Header Functional Descriptor is 0x00
		0x10, 0x01, // bcdCDC field, CDC version is 0x0110


		0x05, // bFunctionLength field, Call Management Functional Descriptor has 5 bytes
		0x24, // bDescriptorType field, CS_INTERFACE is 0x24
		0x01, // bDescriptorSubtype field, Call Management functional descriptor is 0x01
		0x00, // bmCapabilities field, we don't manage call management
		0x00, // bDataInterface field, no data interface is used for call management


		0x04, // bFunctionLength field, Abstract Control Management Functional Descriptor has 4 bytes
		0x24, // bDescriptorType field, CS_INTERFACE is 0x24
		0x02, // bDescriptorSubtype field, Abstract Control Management functional descriptor is 0x02
		0x02, // bmCapabilities field, we support Set_Line_Coding, Set_Control_Line_State, Get_Line_Coding, and Serial_State


		0x05, // bFunctionLength field, Union Functional Descriptor has 5 bytes
		0x24, // bDescriptorType field, CS_INTERFACE is 0x24
		0x06, // bDescriptorSubtype field, Union functional descriptor is 0x06
		0x00, // MasterInterface field, we need the number 0 CDC interface here
		0x01, // SlaveInterface, we need the number 1 data interface here


		0x07, // bLength field, end point descriptor has 7 bytes
		0x05, // bDescriptorType field, end point descriptor is 0x05
		// bEndpointAddress field, we use end point 3 here
		0x83, // the MSb indicates this is an input end point
		0x03, // bmAttributes field, interrupt mode is 3
		// wMaxPacketSize field, end point 3 has 64 bytes capacity
		0x40, 0x00,
		// bInterval field, we use 1 frames here, which is 1ms
		0x01,

		0x09, // bLength field, interface descriptor has 9 bytes
		0x04, // bDescriptorType field, interface descriptor is 0x04
		0x01, // bInterfaceNumber field, this interface is number 1
		0x00, // bAlternateSetting field
		0x02, // bNumEndpoints field, we use 2 end point other than EP0
		0x0A, // bInterfaceClass field, data interface is 0x0A
		0x00, // bInterfaceSubClass field
		0x00, // bInterfaceProtocol field
		0x00, // iConfiguration field


		0x07, // bLength field, end point descriptor has 7 bytes
		0x05, // bDescriptorType field, end point descriptor is 0x05
		// bEndpointAddress field, we use end point 1 here
		0x81, // the MSb indicates this is an input end point
		0x02, // bmAttributes field, bulk mode is 0x02
		0x40, 0x00, // wMaxPacketSize field, end point 1 has 64 bytes capacity
		0x00, // bInterval field


		0x07, // bLength field, end point descriptor has 7 bytes
		0x05, // bDescriptorType field, end point descriptor is 0x05
		// bEndpointAddress field, we use end point 2 here
		0x02, // the MSb indicates this is an output end point
		0x02, // bmAttributes field, bulk mode is 0x02
		0x20, 0x00, // wMaxPacketSize field, end point 2 has 32 bytes capacity
		0x00, // bInterval field
		};

const uint8_t language_id[4] = { 0x04, // bLength field, this string descriptor has 4 bytes
		0x03, // bDescriptorType, string descriptor is 0x03
		0x09, 0x04 // language ID, en_us is 0x0409
		};

const uint8_t manufacturer_string_descriptor[36] = {
		0x24, // bLength field, this string descriptor has 36 bytes
		0x03, // bDescriptorType, string descriptor is 0x03
		'A', 0x00, 'n', 0x00, 'd', 0x00, 'y', 0x00, 'g', 0x00, 'o', 0x00, 'n',
		0x00, 'g', 0x00, 'y', 0x00, 'b', 0x00, ' ', 0x00, 'S', 0x00, 't', 0x00,
		'u', 0x00, 'd', 0x00, 'i', 0x00, 'o', 0x00, };

const uint8_t product_string_descriptor[58] = {
		0x3A, // bLength field, this string descriptor has 58 bytes
		0x03, // bDescriptorType, string descriptor is 0x03
		'F', 0x00, 'u', 0x00, 'n', 0x00, 'n', 0x00, 'y', 0x00, 'R', 0x00, 'o',
		0x00, 'b', 0x00, 'o', 0x00, 't', 0x00, ' ', 0x00, 'W', 0x00, 'i', 0x00,
		'r', 0x00, 'e', 0x00, 'l', 0x00, 'e', 0x00, 's', 0x00, 's', 0x00, ' ',
		0x00, 'R', 0x00, 'e', 0x00, 'c', 0x00, 'e', 0x00, 'i', 0x00, 'v', 0x00,
		'e', 0x00, 'r', 0x00, };

const uint8_t serial_number_string_descriptor[22] = {
		0x16, // bLength field, this string descriptor has 22 bytes
		0x03, // bDescriptorType, string descriptor is 0x03
		'2', 0x00, '0', 0x00, '1', 0x00, '1', 0x00, '-', 0x00, '1', 0x00, '1',
		0x00, '-', 0x00, '3', 0x00, '0', 0x00, };

const uint8_t GET_STATUS = 0;
const uint8_t CLEAR_FEATURE = 1;
const uint8_t SET_FEATURE = 3;
const uint8_t SET_ADDRESS = 5;
const uint8_t GET_DESCRIPTOR = 6;
const uint8_t SET_DESCRIPTOR = 7;
const uint8_t GET_CONFIGURATION = 8;
const uint8_t SET_CONFIGURATION = 9;
const uint8_t GET_INTERFACE = 10;
const uint8_t SET_INTERFACE = 11;
const uint8_t SYNCH_FRAME = 12;

const uint8_t DEVICE_DESCRIPTOR = 1;
const uint8_t CONFIGURATION_DESCRIPTOR = 2;
const uint8_t STRING_DESCRIPTOR = 3;
const uint8_t INTERFACE_DESCRIPTOR = 4;
const uint8_t ENDPOINT_DESCRIPTOR = 5;
const uint8_t REPORT_DESCRIPTOR = 0x22;

const uint8_t SET_IDLE = 0x0A;

const uint8_t GET_LINE_CODING = 0x21;
const uint8_t SERIAL_STATE = 0x20;
const uint8_t SET_LINE_CODING = 0x20;
const uint8_t SET_CONTROL_LINE_STATE = 0x22;
const uint8_t SEND_BREAK = 0x23;

extern uint8_t line_coding_dummy[7];

}

typedef class CUSB_CDC {
private:

	static CUSB_CDC *m_instance;

	typedef enum {
		DETACHED, ATTACHED, ASSIGN_ADDRESS, ADDRESSED, CONFIGURED, SUSPENDED,
	} state_t;

	typedef enum {
		IDLE, IN, OUT, SET_LINE,
	} request_state_t;

	state_t m_state;
	request_state_t m_request_state;

	typedef union {
		struct {
			struct {
				uint8_t receiver :5;
				uint8_t type :2;
				uint8_t dir :1;
			} request_type;
			uint8_t request;
			uint16_t value;
			uint16_t index;
			uint16_t len;
		};
		uint8_t buffer[8];
	} request_t;

	request_t m_request;

	volatile bool m_tx_empty;
	volatile bool m_if_connected;
	uint8_t m_rx_buffer[128];
	uint8_t m_rx_count;

	CUSB_CDC(void);

	inline void set_address(uint8_t addr);
	inline void enable_endpoints(void);
	inline void enable_endpoint_0(void);

	void attach(void) {
		UDCON &= ~_BV(DETACH);
	}

	void detach(void) {
		UDCON |= _BV(DETACH);
		m_tx_empty = false;
		m_rx_count = 0;
		m_if_connected = false;
	}

	void reset_endpoints(void) {
		UERST = 0x7F; // reset all endpoints
		UERST = 0x00; // finish resetting
	}

	inline int8_t read_request(void);
	inline uint8_t respond_request(const uint8_t *buffer, uint8_t len);

public:

	static CUSB_CDC *instance(void) {
		if (m_instance == NULL) {
			m_instance = new CUSB_CDC();
		}

		return m_instance;
	}

	void pwr(bool if_on);

	uint8_t write_bank(uint8_t *buff, uint8_t num);
	void send(void);

	bool connection(void) {
		return m_if_connected;
	}

	uint8_t read_bank(uint8_t *buff);

	bool ready_to_send(void) {
		return m_tx_empty;
	}

	friend void USB_GEN_vect(void);
	friend void USB_COM_vect(void);

}*pCUSB_CDC;

#endif /* USB_HPP_ */
